"""
Test cases for different tools
"""

from __future__ import print_function

import tempfile
import _setup_test_env  # noqa
from subprocess import Popen, PIPE, STDOUT, CalledProcessError
import os
import sys


my_dir = os.path.dirname(os.path.abspath(__file__))
base_dir = os.path.dirname(my_dir)
py = sys.executable
print("Python:", py)
_run_count = 0


def run(*args):
  args = list(args)
  print("run:", args)
  global _run_count
  if _run_count == 0:
    # For the first run, as a special case, directly run the script in the current env.
    # This is easier for debugging.
    from returnn.util.basic import generic_import_module
    mod = generic_import_module(os.path.join(base_dir, args[0]))
    # noinspection PyUnresolvedReferences
    mod.main(args)
    _run_count += 1
    return
  _run_count += 1
  # RETURNN by default outputs on stderr, so just merge both together
  p = Popen(args, stdout=PIPE, stderr=STDOUT, cwd=base_dir)
  out, _ = p.communicate()
  if p.returncode != 0:
    print("Return code is %i" % p.returncode)
    print("std out/err:\n---\n%s\n---\n" % out.decode("utf8"))
    raise CalledProcessError(cmd=args, returncode=p.returncode, output=out)
  return out.decode("utf8")


###############################
# Tests for compile_tf_graph.py
###############################

rec_encoder_decoder_simple_config = """
#!rnn.py
network = {
    "enc0": {"class": "linear", "from": "data", "activation": "sigmoid", "n_out": 3},
    "enc1": {"class": "reduce", "mode": "max", "axis": "t", "from": "enc0"},
    "output": {
      "class": "rec", "from": [], "target": "classes",
      "unit": {
        "embed": {"class": "linear", "from": "prev:output", "activation": "sigmoid", "n_out": 3},
        "s": {"class": "rec", "unit": "lstm", "from": ["embed", "base:enc1"], "n_out": 3},
        "prob": {"class": "softmax", "from": "s", "loss": "ce", "target": "classes"},
        "output": {"class": "choice", "beam_size": 4, "from": "prob", "target": "classes", "initial_output": 0},
        "end": {"class": "compare", "from": "output", "value": 0}
      }
    },
    "decision": {"class": "decide", "from": "output", "loss": "edit_distance"}
}
num_inputs = 5
num_outputs = 3
use_tensorflow = True
"""


rec_encoder_decoder_att_config = """
#!rnn.py
network = {
    "encoder": {"class": "linear", "from": "data", "activation": "sigmoid", "n_out": 3},
    "enc_ctx": {"class": "linear", "from": "encoder", "n_out": 10},
    "inv_fertility": {"class": "linear", "activation": "sigmoid", "with_bias": False, "from": "encoder", "n_out": 1},
    "output": {
      "class": "rec", "from": [], "target": "classes",
      "unit": {
        "weight_feedback": {"class": "linear", "activation": None, "with_bias": False, "from": "prev:accum_att_weights", "n_out": 10},
        "prev_s_transformed": {"class": "linear", "activation": None, "with_bias": False, "from": "prev:s", "n_out": 10},
        "energy_in": {"class": "combine", "kind": "add", "from": ["base:enc_ctx", "weight_feedback", "prev_s_transformed"], "n_out": 10},
        "energy_tanh": {"class": "activation", "activation": "tanh", "from": "energy_in"},
        "energy": {"class": "linear", "activation": None, "with_bias": False, "from": "energy_tanh", "n_out": 1},
        "att_weights": {"class": "softmax_over_spatial", "from": "energy"},
        "accum_att_weights": {
          "class": "eval", "from": ["prev:accum_att_weights", "att_weights", "base:inv_fertility"],
          "eval": "source(0) + source(1) * source(2) * 0.5",
          "out_type": {"dim": 1, "shape": (None, 1)}},
        "att": {"class": "generic_attention", "weights": "att_weights", "base": "base:encoder", "auto_squeeze": True},

        "s": {"class": "rec", "unit": "lstm", "from": ["att", "prev:embed"], "n_out": 3},
        "prob": {"class": "softmax", "from": "s", "loss": "ce", "target": "classes"},
        "output": {"class": "choice", "beam_size": 4, "from": "prob", "target": "classes", "initial_output": 0},
        "embed": {"class": "linear", "from": "prev:output", "activation": "sigmoid", "n_out": 3},
        "end": {"class": "compare", "from": "output", "value": 0}
      }
    },
    "decision": {"class": "decide", "from": "output", "loss": "edit_distance"}
}
num_inputs = 5
num_outputs = 3
use_tensorflow = True
"""


rec_transducer_time_sync_config = """
#!rnn.py
network = {
    "encoder": {"class": "linear", "from": "data", "activation": "sigmoid", "n_out": 3},
    "output": {
      "class": "rec", "from": "encoder", "target": "classes",
      "unit": {
        "embed": {"class": "linear", "from": "prev:output", "activation": "sigmoid", "n_out": 3},
        "s": {"class": "rec", "unit": "lstm", "from": ["embed", "data:source"], "n_out": 3},
        "prob": {"class": "softmax", "from": "s", "loss": "ce", "target": "classes"},
        "output": {"class": "choice", "beam_size": 4, "from": "prob", "target": "classes", "initial_output": 0},
      }
    },
}
num_inputs = 5
num_outputs = 3
use_tensorflow = True
"""


rec_transducer_time_sync_delayed_config = """
#!rnn.py
network = {
    "encoder": {"class": "linear", "from": "data", "activation": "sigmoid", "n_out": 3},
    "output": {
      "class": "rec", "from": "encoder", "target": "classes",
      "unit": {
        "s": {"class": "rec", "unit": "lstm", "from": ["prev:embed", "prev:s2", "data:source"], "n_out": 3},
        "prob": {"class": "softmax", "from": "s", "loss": "ce", "target": "classes"},
        "output": {"class": "choice", "beam_size": 4, "from": "prob", "target": "classes", "initial_output": 0},
        "embed": {"class": "linear", "from": "output", "activation": "sigmoid", "n_out": 3},
        "s2": {"class": "rec", "unit": "lstm", "from": ["embed", "prev:s"], "n_out": 3},
      }
    },
}
num_inputs = 5
num_outputs = 3
use_tensorflow = True
"""


def test_compile_tf_graph_basic():
  tmp_dir = tempfile.mkdtemp()
  with open(os.path.join(tmp_dir, "returnn.config"), "wt") as config:
    config.write(rec_encoder_decoder_simple_config)
  args = [
    "tools/compile_tf_graph.py",
    "--output_file",
    os.path.join(tmp_dir, "graph.metatxt"),
    os.path.join(tmp_dir, "returnn.config")
  ]
  run(*args)


def test_compile_tf_graph_basic_second_run():
  # Just to make sure that the second run works as well,
  # which behaves different due to the debug case of the first run.
  # See :func:`run` above.
  test_compile_tf_graph_basic()


def test_compile_tf_graph_enc_dec_simple_recurrent_step():
  tmp_dir = tempfile.mkdtemp()
  with open(os.path.join(tmp_dir, "returnn.config"), "wt") as config:
    config.write(rec_encoder_decoder_simple_config)
  args = [
    "tools/compile_tf_graph.py",
    "--output_file",
    os.path.join(tmp_dir, "graph.metatxt"),
    "--rec_step_by_step",
    "output",
    os.path.join(tmp_dir, "returnn.config")
  ]
  run(*args)


def test_compile_tf_graph_enc_dec_att_recurrent_step():
  # https://github.com/rwth-i6/returnn/issues/1016
  tmp_dir = tempfile.mkdtemp()
  with open(os.path.join(tmp_dir, "returnn.config"), "wt") as config:
    config.write(rec_encoder_decoder_att_config)
  args = [
    "tools/compile_tf_graph.py",
    "--output_file",
    os.path.join(tmp_dir, "graph.metatxt"),
    "--rec_step_by_step",
    "output",
    os.path.join(tmp_dir, "returnn.config")
  ]
  run(*args)


def test_compile_tf_graph_transducer_time_sync_recurrent_step():
  tmp_dir = tempfile.mkdtemp()
  with open(os.path.join(tmp_dir, "returnn.config"), "wt") as config:
    config.write(rec_transducer_time_sync_config)
  args = [
    "tools/compile_tf_graph.py",
    "--output_file",
    os.path.join(tmp_dir, "graph.metatxt"),
    "--rec_step_by_step",
    "output",
    os.path.join(tmp_dir, "returnn.config")
  ]
  run(*args)


def test_compile_tf_graph_transducer_time_sync_delayed_recurrent_step():
  tmp_dir = tempfile.mkdtemp()
  with open(os.path.join(tmp_dir, "returnn.config"), "wt") as config:
    config.write(rec_transducer_time_sync_delayed_config)
  args = [
    "tools/compile_tf_graph.py",
    "--output_file",
    os.path.join(tmp_dir, "graph.metatxt"),
    "--rec_step_by_step",
    "output",
    os.path.join(tmp_dir, "returnn.config")
  ]
  run(*args)



"""
Main engine for PyTorch
"""

from torch.utils.data import DataLoader

from returnn.log import log
from returnn.engine.base import EngineBase
from returnn.datasets.basic import init_dataset
from returnn.torch.dataset_wrapper import DatasetWrapper


class Engine(EngineBase):
  """
  PyTorch engine
  """

  def __init__(self, config):
    """
    :param returnn.config.Config config:
    """
    super(Engine, self).__init__()
    self.config = config
    self.train_dataset = None
    self.eval_datasets = {}

  def init_train_from_config(self, config=None, train_data=None, dev_data=None, eval_data=None):
    """
    :param returnn.config.Config|None config:
    :param returnn.datasets.basic.Dataset|None train_data:
    :param returnn.datasets.basic.Dataset|None dev_data:
    :param returnn.datasets.basic.Dataset|None eval_data:
    """
    assert config is self.config
    self.train_dataset = train_data
    self.eval_datasets.clear()
    if dev_data:
      self.eval_datasets["dev"] = dev_data
    if eval_data:
      self.eval_datasets["eval"] = eval_data
    if config.has("eval_datasets"):
      for dataset_name, dataset_opts in config.typed_value("eval_datasets", {}).items():
        self.eval_datasets[dataset_name] = init_dataset(dataset_opts, default_kwargs={"name": dataset_name})

  def train(self):
    """
    Main training loop.
    """
    start_epoch, _ = self.get_train_start_epoch_batch(self.config)
    final_epoch = self.config_get_final_epoch(self.config)

    print("Starting training at epoch {}.".format(start_epoch), file=log.v3)

    self.epoch = start_epoch
    while self.epoch <= final_epoch:
      print("Starting " + self.get_epoch_str() + "...", file=log.v4)

      train_data = DatasetWrapper(self.train_dataset, epoch=self.epoch)

      data_loader = DataLoader(train_data, batch_size=1)  # TODO: implement batching

      for data in data_loader:
        assert data  # TODO: only iterates through dataset so far

      self.epoch += 1

    print("Finished training at epoch {}.".format(self.epoch), file=log.v3)



#!/usr/bin/env python3

"""
Small demo for Horovod + MPI.
"""

import os
import sys


def main():
  """
  Main entry.
  """
  print("pid %i: Hello" % os.getpid())
  print("Python version:", sys.version)

  print("Env:")
  for key, value in sorted(os.environ.items()):
    print("%s=%s" % (key, value))
  print()

  if os.environ.get("PE_HOSTFILE", ""):
    try:
      print("PE_HOSTFILE, %s:" % os.environ["PE_HOSTFILE"])
      with open(os.environ["PE_HOSTFILE"], "r") as f:
        print(f.read())
    except FileNotFoundError as exc:
      print(exc)

  if os.environ.get("SGE_JOB_SPOOL_DIR", ""):
    print("SGE_JOB_SPOOL_DIR, %s:" % os.environ["SGE_JOB_SPOOL_DIR"])
    for name in os.listdir(os.environ["SGE_JOB_SPOOL_DIR"]):
      print(name)
    print()

  if os.environ.get("OMPI_FILE_LOCATION", ""):
    print("OMPI_FILE_LOCATION, %s:" % os.environ["OMPI_FILE_LOCATION"])
    d = os.path.dirname(os.path.dirname(os.environ["OMPI_FILE_LOCATION"]))
    print("dir:", d)
    for name in os.listdir(d):
      print(name)
    print()
    print("contact.txt:")
    with open("%s/contact.txt" % d, "r") as f:
      print(f.read())
    print()

  # https://github.com/horovod/horovod/issues/1123
  try:
    import ctypes
    ctypes.CDLL("libhwloc.so", mode=ctypes.RTLD_GLOBAL)
  except Exception as exc:
    print("Exception while loading libhwloc.so, ignoring...", exc)

  print("sys.path:")
  i = 0
  for p in list(sys.path):
    print(p)
    if "/.local/lib/" in p:
      # small workaround if the order is messed up... prefer from .local/lib
      print("(insert at position %i)" % i)
      sys.path.insert(i, p)
      i += 1
  print()

  try:
    from mpi4py import MPI  # noqa
    name = MPI.Get_processor_name()
    comm = MPI.COMM_WORLD
    print("mpi4py:", "name: %s," % name, "rank: %i," % comm.Get_rank(), "size: %i" % comm.Get_size())
    hosts = comm.allgather((comm.Get_rank(), name))  # Get the names of all the other hosts
    print("  all hosts:", {key: item for (key, item) in hosts})
  except ImportError:
    print("mpi4py not available")

  print("Import TF now...")
  import tensorflow as tf
  print("TF version:", tf.__version__)

  import horovod  # noqa
  print("Horovod version:", horovod.__version__)
  import horovod.tensorflow as hvd  # noqa

  # Initialize Horovod
  hvd.init()

  print(
    "pid %i: hvd: rank: %i, size: %i, local_rank %i, local_size %i" % (
      os.getpid(), hvd.rank(), hvd.size(), hvd.local_rank(), hvd.local_size()))


if __name__ == "__main__":
  main()



"""
Code to create PyTorch datasets that can be used with the PyTorch DataLoader.
"""

from torch.utils.data import IterableDataset


# noinspection PyAbstractClass
class DatasetWrapper(IterableDataset):
  """
  Converts a RETURNN dataset into a PyTorch IterableDataset.
  """

  def __init__(self, returnn_dataset, epoch):
    """
    :param returnn.datasets.basic.Dataset returnn_dataset: dataset to be wrapped
    :param int epoch:
    """
    self._dataset = returnn_dataset
    self._epoch = epoch

  def __iter__(self):
    """
    :return: generator providing data samples in the form of a dict data_key -> data
    :rtype: Iterable[dict[str, numpy.ndarray]]
    """
    self._dataset.init_seq_order(epoch=self._epoch)

    data_keys = self._dataset.get_data_keys()

    seq_index = 0
    while self._dataset.is_less_than_num_seqs(seq_index):
      self._dataset.load_seqs(seq_index, seq_index + 1)

      data = {data_key: self._dataset.get_data(seq_index, data_key) for data_key in data_keys}

      yield data

      seq_index += 1



#!/usr/bin/env python3

# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""
This is mostly a copy of:
https://github.com/tensorflow/tensorflow/blob/master/tensorflow/python/tools/inspect_checkpoint.py

A simple script for inspect checkpoint files.

We extended it a bit.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import _setup_returnn_env  # noqa
import sys
import argparse
import numpy

from tensorflow.python.platform import app
from tensorflow.python.platform import flags
import returnn.tf.compat as tf_compat

FLAGS = None


def print_tensor(v):
  """
  :param numpy.ndarray v:
  """
  print(v)
  # See :func:`variable_scalar_summaries_dict`.
  mean = numpy.mean(v)
  print("mean:", mean)
  print("stddev:", numpy.sqrt(numpy.mean(numpy.square(v - mean))))
  print("rms:", numpy.sqrt(numpy.mean(numpy.square(v))))
  print("min:", numpy.min(v))
  print("max:", numpy.max(v))


def print_tensors_in_checkpoint_file(file_name, tensor_name, all_tensors):
  """Prints tensors in a checkpoint file.

  If no `tensor_name` is provided, prints the tensor names and shapes
  in the checkpoint file.

  If `tensor_name` is provided, prints the content of the tensor.

  Args:
    file_name: Name of the checkpoint file.
    tensor_name: Name of the tensor in the checkpoint file to print.
    all_tensors: Boolean indicating whether to print all tensors.
  """
  try:
    reader = tf_compat.v1.train.NewCheckpointReader(file_name)
    if all_tensors:
      var_to_shape_map = reader.get_variable_to_shape_map()
      for key in sorted(var_to_shape_map):
        print("tensor_name: ", key)
        print(reader.get_tensor(key))
    elif not tensor_name:
      print(reader.debug_string().decode("utf-8"))
    else:
      print("tensor_name: ", tensor_name)
      v = reader.get_tensor(tensor_name)
      print_tensor(v)
      if tensor_name.endswith("/Adam") and reader.has_tensor(tensor_name + "_1"):
        # https://github.com/tensorflow/tensorflow/blob/03beb65cecbc1e49ea477bca7f54543134b31d53/tensorflow/core/kernels/training_ops_gpu.cu.cc
        print("Guessing Adam m/v")
        v2 = reader.get_tensor(tensor_name + "_1")
        eps = 1e-8
        print("Adam update (m / (eps + sqrt(v))) with eps=%r:" % eps)
        print_tensor(v / (eps + numpy.sqrt(v2)))
  except Exception as e:  # pylint: disable=broad-except
    print(str(e))
    if "corrupted compressed block contents" in str(e):
      print("It's likely that your checkpoint file has been compressed with SNAPPY.")
    if "Data loss" in str(e) and (any([e in file_name for e in [".index", ".meta", ".data"]])):
      proposed_file = ".".join(file_name.split(".")[0:-1])
      v2_file_error_template = """
It's likely that this is a V2 checkpoint and you need to provide the filename
*prefix*.  Try removing the '.' and extension.  Try:
inspect checkpoint --file_name = {}"""
      print(v2_file_error_template.format(proposed_file))


def parse_numpy_printoption(kv_str):
  """Sets a single numpy printoption from a string of the form 'x=y'.

  See documentation on numpy.set_printoptions() for details about what values
  x and y can take. x can be any option listed there other than 'formatter'.

  Args:
    kv_str: A string of the form 'x=y', such as 'threshold=100000'

  Raises:
    argparse.ArgumentTypeError: If the string couldn't be used to set any
        nump printoption.
  """
  k_v_str = kv_str.split("=", 1)
  if len(k_v_str) != 2 or not k_v_str[0]:
    raise argparse.ArgumentTypeError("'%s' is not in the form k=v." % kv_str)
  k, v_str = k_v_str
  printoptions = numpy.get_printoptions()
  if k not in printoptions:
    raise argparse.ArgumentTypeError("'%s' is not a valid printoption." % k)
  v_type = type(printoptions[k])
  if printoptions[k] is None:
    raise argparse.ArgumentTypeError(
      "Setting '%s' from the command line is not supported." % k)
  try:
    v = v_type(v_str) if v_type is not bool else flags.BooleanParser().parse(v_str)
  except ValueError as e:
    raise argparse.ArgumentTypeError(str(e))
  numpy.set_printoptions(**{k: v})


# noinspection PyUnusedLocal
def main(unused_argv):
  """
  Main entry:
  """
  if not FLAGS.file_name:
    print(
      "Usage: inspect_checkpoint --file_name=checkpoint_file_name "
      "[--tensor_name=tensor_to_print]")
    sys.exit(1)
  else:
    print_tensors_in_checkpoint_file(FLAGS.file_name, FLAGS.tensor_name, FLAGS.all_tensors)


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.register("type", "bool", lambda v: v.lower() == "true")
  parser.add_argument(
    "--file_name", type=str, default="", help="Checkpoint filename. "
    "Note, if using Checkpoint V2 format, file_name is the "
    "shared prefix between all files in the checkpoint.")
  parser.add_argument(
    "--tensor_name",
    type=str, default="",
    help="Name of the tensor to inspect")
  parser.add_argument(
    "--all_tensors",
    nargs="?",
    const=True,
    type=bool,
    default=False,
    help="If True, print the values of all the tensors.")
  parser.add_argument(
    "--printoptions",
    nargs="*",
    type=parse_numpy_printoption,
    help="Argument for numpy.set_printoptions(), in the form 'k=v'.")
  FLAGS, unparsed = parser.parse_known_args()
  app.run(main=main, argv=[sys.argv[0]] + unparsed)
